package fl.camera

import android.Manifest
import android.annotation.SuppressLint
import android.app.Activity
import android.content.pm.PackageManager
import android.util.Size
import android.view.Surface
import androidx.camera.core.*
import androidx.camera.lifecycle.ProcessCameraProvider
import androidx.core.content.ContextCompat
import androidx.lifecycle.LifecycleOwner
import io.flutter.plugin.common.MethodChannel
import io.flutter.view.TextureRegistry

class FlCameraX(
    private val activity: Activity,
    private val textureRegistry: TextureRegistry,
    private val flCameraEvent: FlCameraEvent
) {

    private var cameraProvider: ProcessCameraProvider? = null
    private var textureEntry: TextureRegistry.SurfaceTextureEntry? = null
    private var camera: Camera? = null

    private fun checkPermission(): Boolean {
        return ContextCompat.checkSelfPermission(
            activity,
            Manifest.permission.CAMERA
        ) == PackageManager.PERMISSION_GRANTED
    }


    @SuppressLint("RestrictedApi")
    fun initCameraX(
        zoom: Int,
        useBackCamera: Boolean,
        result: MethodChannel.Result,
        imageAnalyzer: ImageAnalysis.Analyzer?
    ) {
        if (!checkPermission()) return
        val displayMetrics = activity.resources.displayMetrics
        val width = displayMetrics.widthPixels / zoom
        val height = displayMetrics.heightPixels / zoom
        val provider = ProcessCameraProvider.getInstance(activity)
        val executor = ContextCompat.getMainExecutor(activity)
        provider.addListener({
            cameraProvider = provider.get()
            textureEntry = textureRegistry.createSurfaceTexture()
            val textureId = textureEntry!!.id()
            val surfaceProvider = Preview.SurfaceProvider { request ->
                val resolution = request.resolution
                val texture = textureEntry!!.surfaceTexture()
                texture.setDefaultBufferSize(resolution.width, resolution.height)
                val surface = Surface(texture)
                request.provideSurface(surface, executor, { })
            }
            val preview =
                Preview.Builder()
                    .setTargetResolution(Size(width, height))
                    .build()
                    .apply { setSurfaceProvider(surfaceProvider) }

            val analysis = ImageAnalysis.Builder()
                .setBackpressureStrategy(ImageAnalysis.STRATEGY_KEEP_ONLY_LATEST)
                .setTargetResolution(Size(width, height))
                .build().apply {
                    if (imageAnalyzer != null) {
                        setAnalyzer(executor, imageAnalyzer)
                    }
                }
            val owner = activity as LifecycleOwner
            var cameraSelector = CameraSelector.DEFAULT_BACK_CAMERA
            if (!useBackCamera) cameraSelector = CameraSelector.DEFAULT_FRONT_CAMERA
            try {
                camera = cameraProvider!!.bindToLifecycle(owner, cameraSelector, preview, analysis)
                camera!!.cameraInfo.torchState.observe(owner, { state ->
                    // TorchState.OFF = 0; TorchState.ON = 1
                    flCameraEvent.sendEvent(mapOf("flash" to (state == TorchState.ON)))
                })
                val resolution = preview.attachedSurfaceResolution!!
                val map: MutableMap<String, Any> = HashMap()
                val portrait = camera!!.cameraInfo.sensorRotationDegrees % 180 == 0
                val w = resolution.width.toDouble()
                val h = resolution.height.toDouble()
                map["textureId"] = textureId
                if (portrait) {
                    map["width"] = w
                    map["height"] = h
                } else {
                    map["width"] = h
                    map["height"] = w
                }
                result.success(map)
            } catch (e: Exception) {
                result.success(null)
            }
        }, executor)
    }

    fun setFlashMode(state: Boolean) {
        camera?.cameraControl?.enableTorch(state)
    }


    fun dispose() {
        val owner = activity as LifecycleOwner
        camera?.cameraInfo?.torchState?.removeObservers(owner)
        cameraProvider?.unbindAll()
        textureEntry?.release()
        camera = null
        textureEntry = null
        cameraProvider = null
    }
}
