import AVFoundation
import CoreMotion
import Flutter
import Foundation

public class FlCamera: NSObject, FlutterTexture, AVCaptureVideoDataOutputSampleBufferDelegate {
    // 链接相机用的
    var _captureSession: AVCaptureSession?
    // 获取相机设备
    var _device: AVCaptureDevice?
    // 视频输出
    var _videoOutput: AVCaptureVideoDataOutput?

    var _latestPixelBuffer: CVPixelBuffer?

    var _flCameraEvent: FlCameraEvent

    var _registry: FlutterTextureRegistry

    var textureId: Int64?

    var _cameraPosition: AVCaptureDevice.Position?

    var _captureOutputCallBack: ((_ buffer: CMSampleBuffer) -> Void)?

    init(_ event: FlCameraEvent, _ registry: FlutterTextureRegistry) {
        _registry = registry
        _flCameraEvent = event
        super.init()
    }

    func initCamera(_ useBackCamera: Bool, _ zoom: Int, _ result: @escaping FlutterResult, _ captureOutputCallBack: ((_ buffer: CMSampleBuffer) -> Void)?) {
        _captureOutputCallBack = captureOutputCallBack
        _cameraPosition = useBackCamera ? AVCaptureDevice.Position.back : .front
        if #available(iOS 10.0, *) {
            _device = AVCaptureDevice.DiscoverySession(deviceTypes: [.builtInWideAngleCamera], mediaType: .video, position: _cameraPosition!).devices.first
        } else {
            _device = AVCaptureDevice.devices(for: .video).filter { $0.position == _cameraPosition }.first
        }
        _device!.addObserver(self, forKeyPath: #keyPath(AVCaptureDevice.torchMode), options: .new, context: nil)
        _captureSession = AVCaptureSession()
        _captureSession!.beginConfiguration()
        var videoInput: AVCaptureInput
        do {
            videoInput = try AVCaptureDeviceInput(device: _device!)
        } catch {
            result(nil)
            return
        }

        _videoOutput = AVCaptureVideoDataOutput()
        _videoOutput!.videoSettings = [kCVPixelBufferPixelFormatTypeKey as String: kCVPixelFormatType_32BGRA]
        _videoOutput!.alwaysDiscardsLateVideoFrames = true
        _videoOutput!.setSampleBufferDelegate(self, queue: .main)

        if _captureSession!.canAddInput(videoInput) {
            _captureSession!.addInput(videoInput)
        }

        if _captureSession!.canAddOutput(_videoOutput!) {
            _captureSession!.addOutput(_videoOutput!)
        }

        for connection in _videoOutput!.connections {
            if connection.isVideoOrientationSupported {
                connection.videoOrientation = .portrait
            }
            if _device?.position == .front, connection.isVideoMirroringSupported {
                connection.isVideoMirrored = true
            }
        }
        let dimensions = CMVideoFormatDescriptionGetDimensions(_device!.activeFormat.formatDescription)
        var _previewSize = CGSize(width: Double(dimensions.height), height: Double(dimensions.width))
        if zoom == 2 {
            _captureSession?.canSetSessionPreset(.hd1280x720)
            _previewSize = CGSize(width: 720, height: 1280)
        }
        _captureSession!.commitConfiguration()
        textureId = _registry.register(self)
        result([
            "textureId": textureId!,
            "width": _previewSize.width,
            "height": _previewSize.height,
        ])
        _captureSession?.startRunning()
    }

    public func captureOutput(_ output: AVCaptureOutput, didOutput sampleBuffer: CMSampleBuffer, from connection: AVCaptureConnection) {
        if output == _videoOutput {
            _latestPixelBuffer = CMSampleBufferGetImageBuffer(sampleBuffer)
            _registry.textureFrameAvailable(textureId!)
            if _captureOutputCallBack != nil {
                _captureOutputCallBack?(sampleBuffer)
            }
        }
    }

    public func imageOrientation() -> UIImage.Orientation {
        switch UIDevice.current.orientation {
        case .portrait:
            return _cameraPosition! == .front ? .leftMirrored : .right
        case .landscapeLeft:
            return _cameraPosition! == .front ? .downMirrored : .up
        case .portraitUpsideDown:
            return _cameraPosition! == .front ? .rightMirrored : .left
        case .landscapeRight:
            return _cameraPosition! == .front ? .upMirrored : .down
        case .faceDown, .faceUp, .unknown:
            return .up
        default:
            return .up
        }
    }

    // 打开/关闭 闪光灯
    func setFlashMode(status: Bool) {
        try? _device!.lockForConfiguration()
        if _device!.hasTorch {
            if status {
                _device!.torchMode = .on
            } else {
                _device!.torchMode = .off
            }
        }
        _device!.unlockForConfiguration()
    }

    override public func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey: Any]?, context: UnsafeMutableRawPointer?) {
        // off = 0; on = 1; auto = 2;
        let state = change?[.newKey] as? Int
        var flash: Bool?
        if state != 2 {
            flash = state == 1
        }
        _flCameraEvent.sendEvent(["flash": flash])
    }

    public func copyPixelBuffer() -> Unmanaged<CVPixelBuffer>? {
        if _latestPixelBuffer == nil {
            return nil
        }
        return Unmanaged<CVPixelBuffer>.passRetained(_latestPixelBuffer!)
    }

    func close() {
        _captureSession!.stopRunning()
        _device?.removeObserver(self, forKeyPath: #keyPath(AVCaptureDevice.torchMode))
        _registry.unregisterTexture(textureId!)
        for intput in _captureSession!.inputs {
            _captureSession!.removeInput(intput)
        }
        for output in _captureSession!.outputs {
            _captureSession!.removeOutput(output)
        }
        _device = nil
        _captureSession = nil
        _videoOutput = nil
        _latestPixelBuffer = nil
    }
}
